@@ -13,27 +13,26 @@ from __future__ import annotations
 import enum
 import logging
 
-from abc import ABC, abstractmethod
-from dataclasses import dataclass, field, fields
-from datetime import datetime, date, timedelta, time, timezone
-from enum import Enum
-from typing import List, Optional, Type, TypeVar, ClassVar
+from dataclasses import dataclass, field
+from datetime import datetime, date, timedelta, time
+from typing import ClassVar, List, Optional, Tuple, Type
 import uuid
 
 import isodate
 
-from jsonschema import validate
-from jsonschema.exceptions import ValidationError
 
 from .schema import (
     feature_store_run_schema,
-    feature_store_schema, feature_set_schema, run_status_field,
+    feature_store_schema, feature_set_schema,
     feature_store_execution_statistics_schema, schedule_state_field, schedule_field, never_schedule, daily_schedule,
     cron_schedule, retry_policy_field, never_retry_policy, fixed_retry_policy, version_target_field, branch_target,
     commit_target, destination_reference_field, topic_destination_reference, table_destination_reference,
     folder_destination_reference, credentials_provider_config_field, commit_schema, attributes_field, file_format_field,
+    source_reference_schema,
 )
-from ..utils import parse_bool, INSTANT_FORMAT
+from ..statistics import SummaryStatistics, ExecutionStatistics, TaskStatistics
+from ..utils import parse_bool, map_opt
+from ..utils.serialisation import AnamlBaseClass, TIMESTAMP_FORMAT, AnamlBaseEnum
 
 logger = logging.getLogger(__name__)
 
@@ -48,101 +47,13 @@ logger = logging.getLogger(__name__)
 # MODEL = anaml-server/common/src/main/scala/io/anaml/common/model
 # PYCLIENT = anaml-server/python-client/src
 
-A = TypeVar('A', bound='AnamlBaseClass')
-
-TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
-
-
-def json_safe(
-    data,
-    datetime_format: Optional[str] = None
-):
-    """Convert values to JSON-safe values."""
-    if datetime_format is None:
-        datetime_format = INSTANT_FORMAT
-    if isinstance(data, str) or isinstance(data, bool) or isinstance(data, float) or isinstance(data, int):
-        return data
-    elif data is None:
-        return None
-    elif isinstance(data, AnamlBaseClass):
-        return data.to_json()
-    elif isinstance(data, enum.Enum):
-        if callable(getattr(data, 'to_json', None)):
-            return data.to_json()
-        else:
-            return data.value
-    elif isinstance(data, uuid.UUID):
-        return str(data)
-    elif isinstance(data, time):
-        return data.strftime("%H:%M:%S")
-    elif isinstance(data, datetime):
-        # NB: datetime <: date, so make sure you don't switch the order here!
-        if (datetime_format == INSTANT_FORMAT):
-            return data.astimezone(timezone.utc).strftime(datetime_format)
-        else:
-            return data.strftime(datetime_format)
-    elif isinstance(data, date):
-        return data.isoformat()
-    elif isinstance(data, dict):
-        return {
-            k: json_safe(v, datetime_format) for k, v in data.items()
-        }
-    elif isinstance(data, list):
-        return [json_safe(v, datetime_format) for v in data]
-    else:
-        # TODO: This is probably an error.
-        return data
-
-
-@dataclass(frozen=True)
-class AnamlBaseClass(ABC):
-    """Base class for data types used in the Anaml API."""
-
-    @classmethod
-    def json_schema(cls) -> Optional[dict]:
-        """JSON Schema for this resource type."""
-        return None
-
-    @classmethod
-    @abstractmethod
-    def from_dict(cls: Type[A], data: dict) -> A:
-        """Construct an instance from a validated dictionary of JSON data."""
-        raise NotImplementedError
-
-    def to_dict(self) -> dict:
-        """Return the object as a dictionary.
-
-        This method should not be recursive, so please don't use :ref:`dataclasses.asdict` here.
-        """
-        return {
-            f.name: getattr(self, f.name) for f in fields(self)
-        }
-
-    def to_json(self) -> dict:
-        """Return a JSON dictionary encoding the object."""
-        return json_safe(self.to_dict())
-
-    @classmethod
-    def from_json(cls: Type[A], d: dict) -> A:
-        """Validate a dictionary of JSON data and parse it into a new instance of this type."""
-        try:
-            schema = cls.json_schema()
-            if schema:
-                validate(d, schema)
-            return cls.from_dict(d)
-        except ValidationError:
-            logger.error("Unable to validate {klass} schema {schema} from dict {data}".format(
-                klass=cls.__name__, schema=cls.json_schema(), data=d
-            ))
-            raise
-
 
 @dataclass(frozen=True)
 class Commit(AnamlBaseClass):
     """An Anaml commit."""
 
     id: uuid.UUID
-    parents: list[uuid.UUID]
+    parents: List[uuid.UUID]
     createdAt: datetime
     author: int
     description: Optional[str]
@@ -169,7 +80,7 @@ class Commit(AnamlBaseClass):
 # Attributes $MODEL/AttributeData.scala
 #
 Label = str
-Attribute = tuple[str, str]  # (key, value)
+Attribute = Tuple[str, str]  # (key, value)
 
 EntityId = int
 EntityName = str
@@ -183,6 +94,16 @@ class Ref:
     ref_type: str = field(init=False, repr=False)
     ref: str
 
+    @classmethod
+    def branch(cls, name: str) -> BranchRef:
+        """Return a branch reference."""
+        return BranchRef(name)
+
+    @classmethod
+    def commit(cls, id: str) -> CommitRef:
+        """Build a commit reference."""
+        return CommitRef(id)
+
 
 @dataclass(frozen=True)
 class CommitRef(Ref):
@@ -207,8 +128,8 @@ class Entity(AnamlBaseClass):
     versionId: EntityVersionId
     description: str
     defaultColumn: str
-    attributes: list[Attribute]
-    labels: list[Label]
+    attributes: List[Attribute]
+    labels: List[Label]
 
     @classmethod
     def json_schema(cls) -> Optional[dict]:
@@ -281,13 +202,13 @@ class EntityMapping(AnamlBaseClass):
 # QualityRating $MODEL/QualityRating.scala
 #
 
-class QualityRating(Enum):
+@enum.unique
+class QualityRating(AnamlBaseEnum):
     """Data asset quality ratings built in to Anaml."""
 
     GOLD = "gold"
     SILVER = "silver"
     BRONZE = "bronze"
-    NULL = None
 
 
 #
@@ -309,8 +230,8 @@ class Source(AnamlBaseClass):
     description: str
     version: EntityVersionId
     predecessor: Optional[SourceVersionId]
-    attributes: list[Attribute] = field(default_factory=list)
-    labels: list[Label] = field(default_factory=list)
+    attributes: List[Attribute] = field(default_factory=list)
+    labels: List[Label] = field(default_factory=list)
 
     @classmethod
     def json_schema(cls) -> Optional[dict]:
@@ -346,7 +267,7 @@ class SourceReference(AnamlBaseClass):
     def json_schema(cls) -> Optional[dict]:
         """JSON schema for source reference objects."""
         # TODO: JSON schema for source reference.
-        return None
+        return source_reference_schema
 
     @classmethod
     def from_dict(cls, data: dict) -> SourceReference:
@@ -512,7 +433,10 @@ class FileFormat(AnamlBaseClass):
     ADT_TYPE: ClassVar[str] = ""
     adt_type: str = field(default=ADT_TYPE, init=False, repr=False)
 
-    # TODO: Refactor in line with new pattern for class-clusters.
+    @property
+    def suffix(self):
+        """Suffix for files in this format."""
+        return f".{self.ADT_TYPE}"
 
     @classmethod
     def json_schema(cls) -> dict:
@@ -522,6 +446,7 @@ class FileFormat(AnamlBaseClass):
     @classmethod
     def from_dict(cls, data: dict) -> FileFormat:
         """Parse a file format from valid JSON data."""
+        # TODO: Refactor in line with new pattern for class-clusters.
         if 'orc' == data['adt_type']:
             return Orc()
         elif 'parquet' == data['adt_type']:
@@ -557,7 +482,7 @@ class Csv(FileFormat):
 
 
 @enum.unique
-class RunStatus(enum.Enum):
+class RunStatus(AnamlBaseEnum):
     """Status of Anaml job runs."""
 
     Pending = 'pending'
@@ -565,26 +490,15 @@ class RunStatus(enum.Enum):
     Completed = 'completed'
     Failed = 'failed'
 
-    @classmethod
-    def json_schema(cls) -> dict:
-        """JSON schema for run status fields."""
-        return run_status_field
-
-    @classmethod
-    def from_dict(cls: Type[RunStatus], data: dict) -> RunStatus:
-        """Parse a run status from valid JSON data."""
-        return cls(data['adt_type'])
-
-    def to_json(self) -> dict:
-        """Convert a run status to JSON data."""
-        return {'adt_type': self.value}
-
 
 @dataclass(frozen=True)
 class FeatureStoreExecutionStatistics(AnamlBaseClass):
     """Statistics from a feature store execution job."""
 
-    # TODO: Complete feature store execution statistics stub.
+    base: ExecutionStatistics
+    featureCount: Optional[int]
+    featureStatistics: List[SummaryStatistics]
+    tasksStatistics: List[TaskStatistics]
 
     @classmethod
     def json_schema(cls) -> dict:
@@ -594,7 +508,12 @@ class FeatureStoreExecutionStatistics(AnamlBaseClass):
     @classmethod
     def from_dict(cls: Type[FeatureStoreExecutionStatistics], data: dict) -> FeatureStoreExecutionStatistics:
         """Parse feature store execution statistics from valid JSON data."""
-        return FeatureStoreExecutionStatistics()
+        return FeatureStoreExecutionStatistics(
+            base=ExecutionStatistics.from_dict(data['base']),
+            featureCount=map_opt(data.get("featureCount", None), int),
+            featureStatistics=[SummaryStatistics.from_dict(s) for s in data['featureStatistics']],
+            tasksStatistics=[TaskStatistics.from_dict(s) for s in data['tasksStatistics']]
+        )
 
 
 @dataclass(frozen=True)
@@ -642,6 +561,7 @@ class FixedRetryPolicy(RetryPolicy):
     """Retry failed jobs according to a fixed backup policy."""
 
     ADT_TYPE: ClassVar[str] = "fixed"
+    adt_type: str = field(init=False, repr=False, default=ADT_TYPE)
 
     backoff: timedelta
     maxAttempts: int
@@ -687,6 +607,7 @@ class DailySchedule(Schedule):
     """Schedule for jobs that should be run once daily."""
 
     ADT_TYPE: ClassVar[str] = "daily"
+    adt_type: str = field(init=False, repr=False, default=ADT_TYPE)
 
     startTimeOfDay: time
 
@@ -751,7 +672,7 @@ class ScheduleState(AnamlBaseClass):
     """State of an Anaml job run."""
 
     schedule: Schedule
-    scheduleStartTime: datetime
+    scheduledStartTime: datetime
     retryCount: int
 
     @classmethod
@@ -764,7 +685,7 @@ class ScheduleState(AnamlBaseClass):
         """Parse a schedule state object from valid JSON data."""
         return ScheduleState(
             schedule=Schedule.from_dict(data['schedule']),
-            scheduleStartTime=isodate.parse_datetime(data['scheduleStartTime']),
+            scheduledStartTime=isodate.parse_datetime(data['scheduledStartTime']),
             retryCount=int(data['retryCount'])
         )
 
@@ -790,20 +711,28 @@ class FeatureStoreRun(AnamlBaseClass):
         """JSON schema for feature store run objects."""
         return feature_store_run_schema
 
+    def to_dict(self) -> dict:
+        """Serialise to a dictionary of JSON data."""
+        return {
+            "id": self.id,
+            "created": self.created.strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
+            "featureStoreId": self.featureStoreId,
+            "featureStoreVersionId": self.featureStoreVersionId,
+            "commitId": self.commitId,
+            "runStartDate": self.runStartDate,
+            "runEndDate": self.runEndDate,
+            "status": self.status,
+            "errorMessage": self.errorMessage,
+            "scheduleState": self.scheduleState,
+            "statistics": self.statistics,
+        }
+
     @classmethod
     def from_dict(cls: Type[FeatureStoreRun], data: dict) -> FeatureStoreRun:
         """Parse a feature store run from valid JSON data."""
-        # TODO: Refactor to new pattern for handling optional fields.
-        schedule_state = data.get('scheduleState', None)
-        if schedule_state:
-            schedule_state = ScheduleState.from_dict(schedule_state)
-        # TODO: Refactor to new pattern for handling optional fields.
-        statistics = data.get('statistics', None)
-        if statistics:
-            statistics = FeatureStoreExecutionStatistics.from_dict(statistics)
         return FeatureStoreRun(
             id=int(data['id']),
-            created=datetime.strptime(data['created'], INSTANT_FORMAT),
+            created=isodate.parse_datetime(data['created']),
             featureStoreId=int(data['featureStoreId']),
             featureStoreVersionId=uuid.UUID(data['featureStoreVersionId']),
             commitId=uuid.UUID(data['commitId']),
@@ -811,8 +740,8 @@ class FeatureStoreRun(AnamlBaseClass):
             runEndDate=date.fromisoformat(data['runEndDate']),
             status=RunStatus.from_dict(data['status']),
             errorMessage=data.get('errorMessage', None),
-            scheduleState=schedule_state,
-            statistics=statistics
+            scheduleState=map_opt(data.get('scheduleState', None), ScheduleState.from_dict),
+            statistics=map_opt(data.get('statistics', None), FeatureStoreExecutionStatistics.from_dict)
         )
 
 
@@ -1036,11 +965,12 @@ class FeatureStore(AnamlBaseClass):
 class FeatureSet(AnamlBaseClass):
     """A collection of related features."""
 
-    id: int
     name: str
     description: str
     features: List[int]
 
+    id: Optional[int] = None
+
     @classmethod
     def json_schema(cls) -> dict:
         """JSON schema for feature set objects."""
